class LocalDlmsProtocolError(Exception):
    """Protocol error"""


class ApplicationAssociationError(Exception):
    """Something went wrong when trying to setup the application association"""


class PreEstablishedAssociationError(Exception):
    """An error when doing illegal things to the connection if it pre established"""


class ConformanceError(Exception):
    """If APDUs does not match connection Conformance"""


class CipheringError(Exception):
    """Something went wrong when ciphering or deciphering an APDU"""


class DlmsClientException(Exception):
    """An exception that is relating to the client"""


class CommunicationError(Exception):
    """Something went wrong in the communication with a meter"""



from functools import partial
from typing import *

import attr

from dlms_cosem import a_xdr, cosem
from dlms_cosem import enumerations as enums
from dlms_cosem.a_xdr import Attribute, AXdrDecoder, EncodingConf
from dlms_cosem.protocol.xdlms import selective_access
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu
from dlms_cosem.protocol.xdlms.invoke_id_and_priority import InvokeIdAndPriority

get_request_type_from_bytes = partial(enums.GetRequestType.from_bytes, byteorder="big")
get_response_type_from_bytes = partial(
    enums.GetResponseType.from_bytes, byteorder="big"
)


class NullValue:
    def __call__(self):
        return None


def if_falsy_set_none(value):
    if value:
        return value


@attr.s(auto_attribs=True)
class GetRequestNormal(AbstractXDlmsApdu):
    """
    Represents a Get request.

    Get requests are modeled with a choice but we only support the normal one.

    Get requests work in single attributes on interface classes.
    To get a value you would need the interface class, the instance (OBIS) and the
    attribute id.

    Some attributes allow for selective access to the attributes. For example a load
    profile might be read from a specific date or a specific entry.
    """

    TAG: ClassVar[int] = 192
    REQUEST_TYPE: ClassVar[enums.GetRequestType] = enums.GetRequestType.NORMAL
    # TODO: try this out
    ENCODING_CONF: ClassVar[EncodingConf] = EncodingConf(
        attributes=[
            Attribute(
                attribute_name="invoke_id_and_priority",
                create_instance=InvokeIdAndPriority.from_bytes,
                length=1,
            ),
            Attribute(
                attribute_name="cosem_attribute",
                create_instance=cosem.CosemAttribute.from_bytes,
                length=9,
            ),
            Attribute(
                attribute_name="access_selection", create_instance=bytes, default=False
            ),
        ]
    )

    cosem_attribute: cosem.CosemAttribute = attr.ib(
        validator=attr.validators.instance_of(cosem.CosemAttribute)
    )
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )
    access_selection: Optional[selective_access.EntryDescriptor] = attr.ib(
        default=None, converter=if_falsy_set_none
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag for GET request is not correct. Got {tag}, should be {cls.TAG}"
            )

        type_choice = enums.GetRequestType(data.pop(0))
        if type_choice is not enums.GetRequestType.NORMAL:
            raise ValueError(
                "The data for the GetRequest is not for a GetRequestNormal"
            )
        decoder = AXdrDecoder(encoding_conf=cls.ENCODING_CONF)
        out_dict = decoder.decode(data)
        return cls(**out_dict)

    def to_bytes(self):
        # automatically adding the choice for GetRequestNormal.
        out = bytearray()
        out.append(self.TAG)
        out.append(self.REQUEST_TYPE.value)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.extend(self.cosem_attribute.to_bytes())

        if self.access_selection:
            out.extend(b"\x01")
            out.extend(self.access_selection.to_bytes())
        else:
            out.extend(b"\x00")

        return bytes(out)


@attr.s(auto_attribs=True)
class GetRequestNext(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 192
    REQUEST_TYPE: ClassVar[enums.GetRequestType] = enums.GetRequestType.NEXT

    block_number: int = attr.ib(validator=attr.validators.instance_of(int), default=0)
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag for GET request is not correct. Got {tag}, should be {cls.TAG}"
            )

        type_choice = enums.GetRequestType(data.pop(0))
        if type_choice is not enums.GetRequestType.NEXT:
            raise ValueError("The data for the GetRequest is not for a GetRequestNext")
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        assert len(data) == 4  # should only be block number left.
        block_number = int.from_bytes(data, "big")
        return cls(block_number, invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.REQUEST_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.extend(self.block_number.to_bytes(4, "big"))
        return bytes(out)


class GetRequestWithList(AbstractXDlmsApdu):
    pass


@attr.s(auto_attribs=True)
class GetRequestFactory:
    """
    The factory will parse the GetRequest and return either a GetRequestNormal,
    GetRequestNext or a GetRequestWithList.
    """

    TAG: ClassVar[int] = 192

    @staticmethod
    def from_bytes(source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != GetRequestFactory.TAG:
            raise ValueError(
                f"Tag for GET request is not correct. Got {tag}, should be "
                f"{GetRequestFactory.TAG}"
            )
        request_type = enums.GetRequestType(data.pop(0))
        if request_type == enums.GetRequestType.NORMAL:
            return GetRequestNormal.from_bytes(source_bytes)
        elif request_type == enums.GetRequestType.NEXT:
            return GetRequestNext.from_bytes(source_bytes)
        elif request_type == enums.GetRequestType.WITH_LIST:
            raise NotImplementedError("GetRequestWithList not implemented")
        else:
            raise ValueError(
                f"Received an enum request type that is not valid for "
                f"GetRequest: {request_type}"
            )


@attr.s(auto_attribs=True)
class GetResponseNormal(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 196
    RESPONSE_TYPE: ClassVar[enums.GetResponseType] = enums.GetResponseType.NORMAL

    data: bytes = attr.ib(validator=attr.validators.instance_of(bytes))
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but is {tag}")
        response_type = enums.GetResponseType(data.pop(0))
        if response_type != cls.RESPONSE_TYPE:
            raise ValueError(
                f"The response type byte: {response_type} is not for a GetResponseNormal"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        choice = data.pop(0)
        if choice != 0:
            raise ValueError(f"The data choice is not 0 to indicate data but: {choice}")

        return cls(bytes(data), invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE)
        out.append(self.invoke_id_and_priority.to_bytes())
        out.append(0)  # data result choice
        out.extend(self.data)
        return bytes(out)


@attr.s(auto_attribs=True)
class GetResponseNormalWithError(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 196
    RESPONSE_TYPE: ClassVar[enums.GetResponseType] = enums.GetResponseType.NORMAL

    error: enums.DataAccessResult = attr.ib(
        validator=attr.validators.instance_of(enums.DataAccessResult)
    )
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but is {tag}")
        response_type = enums.GetResponseType(data.pop(0))
        if response_type != cls.RESPONSE_TYPE:
            raise ValueError(
                f"The response type byte: {response_type} is not for a GetResponseNormal"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        choice = data.pop(0)
        if choice != 1:
            raise ValueError(
                f"The data choice is not 1 to indicate error but: {choice}"
            )

        error = enums.DataAccessResult(data.pop(0))

        return cls(error, invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(1)  # data error choice
        out.extend(self.error.to_bytes(1, "big"))
        return bytes(out)


@attr.s(auto_attribs=True)
class GetResponseWithBlock(AbstractXDlmsApdu):
    """
    The data sent in a block response is an OCTET STRING. Not instance of DLMS Data.
    So it has the length encoding first.

    """

    TAG: ClassVar[int] = 196
    RESPONSE_TYPE: ClassVar[enums.GetResponseType] = enums.GetResponseType.WITH_BLOCK
    data: bytes = attr.ib(validator=attr.validators.instance_of(bytes))
    block_number: int = attr.ib(validator=attr.validators.instance_of(int), default=0)
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but is {tag}")
        response_type = enums.GetResponseType(data.pop(0))
        if response_type != cls.RESPONSE_TYPE:
            raise ValueError(
                f"The response type byte: {response_type} is not for a GetResponseNormal"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        last_block = bool(data.pop(0))
        if last_block:
            raise ValueError(
                f"Last block set to true in a GetResponseWithBlock. Should only be set "
                f"for a GetResponseLastBlock"
            )
        block_number = int.from_bytes(data[:4], "big")

        data = data[4:]
        choice = data.pop(0)
        if choice != 0:
            raise ValueError(f"The data choice is not 0 to indicate data but: {choice}")

        data_length, data = a_xdr.decode_variable_integer(data)
        if data_length != len(data):
            raise ValueError(
                "The octet string in block data is not of the correct length"
            )

        return cls(bytes(data), block_number, invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(0)  # last block == False
        out.extend(self.block_number.to_bytes(4, "big"))
        out.append(0)  # data choice = data
        out.extend(a_xdr.encode_variable_integer(len(self.data)))  # octet string length
        out.extend(self.data)

        return bytes(out)


@attr.s(auto_attribs=True)
class GetResponseLastBlock(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 196
    RESPONSE_TYPE: ClassVar[enums.GetResponseType] = enums.GetResponseType.WITH_BLOCK
    data: bytes = attr.ib(validator=attr.validators.instance_of(bytes))
    block_number: int = attr.ib(validator=attr.validators.instance_of(int), default=0)
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but is {tag}")
        response_type = enums.GetResponseType(data.pop(0))
        if response_type != cls.RESPONSE_TYPE:
            raise ValueError(
                f"The response type byte: {response_type} is not for a GetResponseNormal"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        last_block = bool(data.pop(0))
        if not last_block:
            raise ValueError(
                f"Last block is not set to true in a GetResponseLastBlock."
            )
        block_number = int.from_bytes(data[:4], "big")
        data = data[4:]
        choice = data.pop(0)
        if choice != 0:
            raise ValueError(f"The data choice is not 0 to indicate data but: {choice}")

        data_length, data = a_xdr.decode_variable_integer(data)
        if data_length != len(data):
            raise ValueError(
                "The octet string in block data is not of the correct length"
            )

        return cls(bytes(data), block_number, invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(1)  # last block == True
        out.extend(self.block_number.to_bytes(4, "big"))
        out.append(0)  # data choice = data
        out.extend(a_xdr.encode_variable_integer(len(self.data)))  # octet string length
        out.extend(self.data)
        return bytes(out)


@attr.s(auto_attribs=True)
class GetResponseLastBlockWithError(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 196
    RESPONSE_TYPE: ClassVar[enums.GetResponseType] = enums.GetResponseType.WITH_BLOCK
    error: enums.DataAccessResult = attr.ib(
        validator=attr.validators.instance_of(enums.DataAccessResult)
    )
    block_number: int = attr.ib(validator=attr.validators.instance_of(int), default=0)
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but is {tag}")
        response_type = enums.GetResponseType(data.pop(0))
        if response_type != cls.RESPONSE_TYPE:
            raise ValueError(
                f"The response type byte: {response_type} is not for a GetResponseNormal"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        last_block = bool(data.pop(0))
        if not last_block:
            raise ValueError(
                f"Last block is not set to true in a GetResponseLastBlock."
            )
        block_number = int.from_bytes(data[:4], "big")
        data = data[4:]
        choice = data.pop(0)
        if choice != 1:
            raise ValueError(
                f"The data choice is not 1 to indicate error but: {choice}"
            )

        assert len(data) == 1
        error = enums.DataAccessResult(data.pop(0))
        return cls(error, block_number, invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(1)  # last block == True
        out.extend(self.block_number.to_bytes(4, "big"))
        out.append(1)  # data choice = error
        out.extend(self.error.to_bytes(1, "big"))
        return bytes(out)


@attr.s(auto_attribs=True)
class GetResponseWithList(AbstractXDlmsApdu):
    pass


@attr.s(auto_attribs=True)
class GetResponseFactory:

    TAG: ClassVar[int] = 196

    @staticmethod
    def from_bytes(source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != GetResponseFactory.TAG:
            raise ValueError(
                f"Tag is not correct. Should be {GetResponseFactory.TAG} but is {tag}"
            )
        response_type = enums.GetResponseType(data.pop(0))
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        if response_type == enums.GetResponseType.NORMAL:
            # check if it is an error or data response by assesing the choice.
            choice = data.pop(0)
            if choice == 0:
                return GetResponseNormal(
                    invoke_id_and_priority=invoke_id_and_priority, data=bytes(data)
                )
            elif choice == 1:
                assert len(data) == 1  # should only be one byte left.
                error = enums.DataAccessResult(data.pop(0))
                return GetResponseNormalWithError(
                    invoke_id_and_priority=invoke_id_and_priority, error=error
                )
        elif response_type == enums.GetResponseType.WITH_BLOCK:
            last_block = bool(data.pop(0))
            block_number = int.from_bytes(data[:4], "big")
            data = data[4:]
            choice = data.pop(0)
            if choice == 0:
                data_length, data = a_xdr.decode_variable_integer(data)
                if data_length != len(data):
                    raise ValueError(
                        "The octet string in block data is not of the correct length"
                    )

                if last_block:
                    return GetResponseLastBlock(
                        bytes(data), block_number, invoke_id_and_priority
                    )
                else:
                    return GetResponseWithBlock(
                        bytes(data), block_number, invoke_id_and_priority
                    )
            elif choice == 1:
                assert len(data) == 1  # should only be one byte left.
                error = enums.DataAccessResult(data.pop(0))
                if last_block:
                    return GetResponseLastBlockWithError(
                        error, block_number, invoke_id_and_priority
                    )
                else:
                    raise ValueError(
                        "It is not possible to send an error on a "
                        "GetResponseWithBlock. When an error occurs it "
                        "should always be sent in a GetResponseLastBlockWithError"
                    )

        elif response_type == enums.GetResponseType.WITH_LIST:
            raise NotImplementedError("GetResponseWithList is not implemented.")



from datetime import datetime
from typing import *

import attr

from dlms_cosem import cosem, dlms_data, time


@attr.s(auto_attribs=True)
class CaptureObject:
    """
    Definition of a value that is supposed to be saved in a Profile Generic.

    A data_index of 0 means the whole attribute is referenced. Otherwise it points to a
    specific element of the attribute. For example and entry in a buffer.
    """

    cosem_attribute: cosem.CosemAttribute
    data_index: int = attr.ib(default=0)

    @classmethod
    def from_bytes(cls, source_bytes):
        raise NotImplementedError()

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.extend(b"\x02\x04")  # A structure of 4 elements
        out.extend(
            dlms_data.UnsignedLongData(self.cosem_attribute.interface.value).to_bytes()
        )
        out.extend(
            dlms_data.OctetStringData(
                self.cosem_attribute.instance.to_bytes()
            ).to_bytes()
        )
        out.extend(dlms_data.IntegerData(self.cosem_attribute.attribute).to_bytes())
        out.extend(dlms_data.UnsignedLongData(self.data_index).to_bytes())
        return bytes(out)


@attr.s(auto_attribs=True)
class RangeDescriptor:
    """
    The range descriptor can be used to read buffers of Profile Generic.
    Only buffer element that corresponds to the descriptor shall be returned in a get
    request.


    """

    ACCESS_DESCRIPTOR: ClassVar[int] = 1

    restricting_object: CaptureObject = attr.ib(
        validator=attr.validators.instance_of(CaptureObject)
    )
    from_value: datetime = attr.ib(validator=attr.validators.instance_of(datetime))
    to_value: datetime = attr.ib(validator=attr.validators.instance_of(datetime))
    # selected_values: List[CaptureObject] = attr.ib(factory=list)

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        raise NotImplementedError()

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.ACCESS_DESCRIPTOR)
        out.extend(b"\x02\x04")  # structure of 4 elements
        out.extend(self.restricting_object.to_bytes())
        out.extend(
            dlms_data.OctetStringData(
                time.datetime_to_bytes(self.from_value)
            ).to_bytes()
        )
        out.extend(
            dlms_data.OctetStringData(time.datetime_to_bytes(self.to_value)).to_bytes()
        )
        out.extend(b"\x01\x00")  # empty array for selected values means all columns
        # TODO: implement selected values
        return bytes(out)


def validate_unsigned_double_long_int(instance, attribute, value):
    if 0 >= value >= 0xFFFFFFFF:
        raise ValueError(
            f"{value} is not withing the limits of a unsigned double long integer"
        )


def validate_unsigned_long_int(instance, attribute, value):
    if 0 >= value >= 0xFFFF:
        raise ValueError(
            f"{value} is not withing the limits of a unsigned long integer"
        )


@attr.s(auto_attribs=True)
class EntryDescriptor:
    """
    The entry descriptor limits response data by entries.
    It is possible to limit the entries and also the columns returned.
    The from/to_selected_value limits the columns returned from/to_entry limits the
    entries.

    Numbering of selected values and entries start from 1.
    Setting to_entry=0 or to_selected_value=0 requests the highest possible value.
    """

    ACCESS_DESCRIPTOR: ClassVar[int] = 2

    from_entry: int = attr.ib(
        validator=[validate_unsigned_double_long_int, attr.validators.instance_of(int)]
    )
    to_entry: int = attr.ib(
        validator=[validate_unsigned_double_long_int, attr.validators.instance_of(int)],
        default=0,
    )
    from_selected_value: int = attr.ib(
        validator=[validate_unsigned_long_int, attr.validators.instance_of(int)],
        default=1,
    )
    to_selected_value: int = attr.ib(
        validator=[validate_unsigned_long_int, attr.validators.instance_of(int)],
        default=0,
    )

    @classmethod
    def from_bytes(cls, source_bytes):
        pass

    def to_bytes(self) -> bytes:
        raise NotImplementedError()



import abc
import datetime
from typing import *

import attr

from dlms_cosem import time

VARIABLE_LENGTH = -1


class AbstractDlmsData(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def from_bytes(cls, source_bytes: bytes):
        pass

    @abc.abstractmethod
    def to_python(self) -> Any:
        pass


@attr.s(auto_attribs=True)
class BaseDlmsData(AbstractDlmsData):
    TAG: ClassVar[int] = 0
    LENGTH: ClassVar[int] = 0

    value: Any

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        raise NotImplementedError(
            f"{cls.__name__} have not implemented byte conversion"
        )

    def to_python(self) -> Any:
        return self.value

    def value_to_bytes(self) -> bytes:
        raise NotImplementedError("value_to_bytes must be implemented in subclass")

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        value_bytes = self.value_to_bytes()
        if self.LENGTH == VARIABLE_LENGTH:
            out.append(len(value_bytes))
        out.extend(value_bytes)
        return bytes(out)


@attr.s(auto_attribs=True)
class NullData(BaseDlmsData):
    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(None)

    def to_python(self) -> Any:
        return None

    TAG = 0


@attr.s(auto_attribs=True)
class DataArray(BaseDlmsData):
    """Sequence of Data"""

    TAG = 1
    LENGTH = VARIABLE_LENGTH


@attr.s(auto_attribs=True)
class DataStructure(BaseDlmsData):
    """SEQUENCE of Data"""

    TAG = 2
    LENGTH = VARIABLE_LENGTH


@attr.s(auto_attribs=True)
class BooleanData(BaseDlmsData):
    TAG = 3
    LENGTH = 1

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        value = bool(int.from_bytes(bytes_data, "big"))
        return cls(value)  # TODO: test this.


@attr.s(auto_attribs=True)
class BitStringData(BaseDlmsData):
    TAG = 4
    LENGTH = VARIABLE_LENGTH


@attr.s(auto_attribs=True)
class DoubleLongData(BaseDlmsData):
    """32 bit integer"""

    TAG = 5
    LENGTH = 4

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big", signed=True))


@attr.s(auto_attribs=True)
class DoubleLongUnsignedData(BaseDlmsData):
    """32 bit unsigned integer"""

    TAG = 6
    LENGTH = 4

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big"))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(4, "big")


@attr.s(auto_attribs=True)
class OctetStringData(BaseDlmsData):
    TAG = 9
    LENGTH = VARIABLE_LENGTH

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=bytes_data)

    def value_to_bytes(self) -> bytes:
        return self.value

    def to_python(self) -> bytes:
        return self.value


@attr.s(auto_attribs=True)
class VisibleStringData(BaseDlmsData):
    TAG = 10
    LENGTH = VARIABLE_LENGTH


@attr.s(auto_attribs=True)
class UTF8StringData(BaseDlmsData):
    TAG = 12
    LENGTH = VARIABLE_LENGTH


@attr.s(auto_attribs=True)
class BCDData(BaseDlmsData):
    TAG = 13
    LENGTH = VARIABLE_LENGTH


@attr.s(auto_attribs=True)
class IntegerData(BaseDlmsData):
    """"8 bit integer"""

    TAG = 15
    LENGTH = 1

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big", signed=True))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(1, "big")


@attr.s(auto_attribs=True)
class LongData(BaseDlmsData):
    """16  bit integer"""

    TAG = 16
    LENGTH = 2

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big", signed=True))


@attr.s(auto_attribs=True)
class UnsignedIntegerData(BaseDlmsData):
    """8 bit unsigned integer"""

    TAG = 17
    LENGTH = 1

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big"))


@attr.s(auto_attribs=True)
class UnsignedLongData(BaseDlmsData):
    """16 bit unsigned integer"""

    TAG = 18
    LENGTH = 2

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big"))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(2, "big")


@attr.s(auto_attribs=True)
class CompactArrayData(BaseDlmsData):
    """
    Contains a Type description and arrray content in form of octet string
    content_description -> Type Description tag = 0
    array_content -> Octet string  tag = 1
    """

    TAG = 19
    LENGTH = VARIABLE_LENGTH


@attr.s(auto_attribs=True)
class Long64Data(BaseDlmsData):
    """
    64 bit integer
    """

    TAG = 20
    LENGTH = 8

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big", signed=True))


@attr.s(auto_attribs=True)
class UnsignedLong64Data(BaseDlmsData):
    """
    64 bit unsigned integer
    """

    TAG = 21
    LENGTH = 8

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big"))


@attr.s(auto_attribs=True)
class EnumData(BaseDlmsData):
    """
    8 bit integer
    """

    TAG = 22
    LENGTH = 1

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big"))


@attr.s(auto_attribs=True)
class Float32Data(BaseDlmsData):
    """
    Octet string of 4 bytes
    """

    TAG = 23
    LENGTH = 4


@attr.s(auto_attribs=True)
class Float64Data(BaseDlmsData):
    """
    Octet string of 8 bytes
    """

    TAG = 24
    LENGTH = 8


@attr.s(auto_attribs=True)
class DateTimeData(BaseDlmsData):
    """Octet string of 12 bytes"""

    TAG = 25
    LENGTH = 12

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        if len(bytes_data) != cls.LENGTH:
            raise ValueError(f"Datetime should be 12 bytes long, got {len(bytes_data)}")
        return cls(time.datetime_from_bytes(bytes_data))


@attr.s(auto_attribs=True)
class DateData(BaseDlmsData):
    """Octet string of 5 bytes"""

    TAG = 26
    LENGTH = 5

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        if len(bytes_data) != cls.LENGTH:
            raise ValueError(f"Date should be 5 bytes long, got {len(bytes_data)}")
        return cls(time.date_from_bytes(bytes_data))


@attr.s(auto_attribs=True)
class TimeData(BaseDlmsData):
    """Octet string of 4 bytes"""

    TAG = 27
    LENGTH = 4

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        if len(bytes_data) != cls.LENGTH:
            raise ValueError(f"Time should be 4 bytes long, got {len(bytes_data)}")
        return cls(time.time_from_bytes(bytes_data))


@attr.s(auto_attribs=True)
class DontCareData(BaseDlmsData):
    """Nulldata"""

    TAG = 255
    LENGTH = 0


@attr.s(auto_attribs=True)
class UnixTimestamp(DoubleLongUnsignedData):
    """Unix timestamps should be represented as double long unsigned"""

    DEFAULT_TIMEZONE = datetime.timezone.utc

    @classmethod
    def from_bytes(cls, bytes_data):
        val = datetime.datetime.fromtimestamp(
            int.from_bytes(bytes_data, "big"), cls.DEFAULT_TIMEZONE
        )
        return cls(value=val)


@attr.s(auto_attribs=True)
class DlmsDataFactory:
    MAP: ClassVar[Dict[int, Type]] = {
        0: NullData,
        1: DataArray,
        2: DataStructure,
        3: BooleanData,
        4: BitStringData,
        5: DoubleLongData,
        6: DoubleLongUnsignedData,
        9: OctetStringData,
        10: VisibleStringData,
        12: UTF8StringData,
        13: BCDData,
        15: IntegerData,
        16: LongData,
        17: UnsignedIntegerData,
        18: UnsignedLongData,
        19: CompactArrayData,
        20: Long64Data,
        21: UnsignedLong64Data,
        22: EnumData,
        23: Float32Data,
        24: Float64Data,
        25: DateTimeData,
        26: DateData,
        27: TimeData,
        255: DontCareData,
    }

    @classmethod
    def get_data_class(cls, tag: int):
        return cls.MAP[tag]



import logging
import os
from typing import *

import attr

from dlms_cosem import enumerations as enums
from dlms_cosem import exceptions, security
from dlms_cosem import state as dlms_state
from dlms_cosem import utils
from dlms_cosem.protocol import acse, xdlms
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu
from dlms_cosem.protocol.xdlms.conformance import Conformance

LOG = logging.getLogger(__name__)


def default_system_title() -> bytes:
    """A non FLAG registed id + 5 random bytes """
    return b"uti" + os.urandom(5)


class XDlmsApduFactory:
    """
    A factory to return the correct APDU depending from the tag.
    """

    APDU_MAP = {
        1: xdlms.InitiateRequestApdu,
        8: xdlms.InitiateResponseApdu,
        14: xdlms.ConfirmedServiceErrorApdu,
        15: xdlms.DataNotificationApdu,
        33: xdlms.GlobalCipherInitiateRequest,
        40: xdlms.GlobalCipherInitiateResponse,
        216: xdlms.ExceptionResponseApdu,
        219: xdlms.GeneralGlobalCipherApdu,
        # ACSE APDUs:
        96: acse.ApplicationAssociationRequestApdu,
        97: acse.ApplicationAssociationResponseApdu,
        98: acse.ReleaseRequestApdu,
        99: acse.ReleaseResponseApdu,
        192: xdlms.GetRequestFactory,
        193: xdlms.SetRequestFactory,
        195: xdlms.ActionRequestFactory,
        196: xdlms.GetResponseFactory,
        197: xdlms.SetResponseFactory,
        199: xdlms.ActionResponseFactory,
    }

    @classmethod
    def apdu_from_bytes(cls, apdu_bytes):
        tag = apdu_bytes[0]

        try:
            apdu_class = cls.APDU_MAP[tag]
        except KeyError as e:
            raise KeyError(f"Tag {tag!r} is not available in DLMS APDU Factory") from e
        return apdu_class.from_bytes(apdu_bytes)


def make_client_to_server_challenge(length: int = 8) -> bytes:
    """
    Return a valid challenge depending on the authentocation method.
    """
    if 8 <= length <= 64:
        return os.urandom(length)
    else:
        raise ValueError(
            f"Client to server challenge must be between 8 and 64 bytes. Got {length}"
        )


def make_conformance(encryption_key: Optional[bytes], use_block_transfer: bool):
    """
    Return a default conformance with general_protection set if a
    encryption key is passed.
    """
    return Conformance(
        general_protection=bool(encryption_key),
        general_block_transfer=True,
        delta_value_encoding=False,
        attribute_0_supported_with_set=False,
        priority_management_supported=True,
        attribute_0_supported_with_get=False,
        block_transfer_with_get_or_read=False,
        block_transfer_with_set_or_write=False,
        block_transfer_with_action=False,
        multiple_references=False,
        data_notification=False,
        access=True,
        get=True,
        set=True,
        selective_access=True,
        event_notification=True,
        action=True,
    )


class ProtectionError(Exception):
    """Unable to perform cryptographic function"""


@attr.s(auto_attribs=True)
class DlmsConnection:
    """
    A DLMS connection.
    """

    # Client system title can be any combination of 8 bytes.
    # But is should not be the same as the metering the connection is set up too.
    client_system_title: bytes = attr.ib(
        converter=attr.converters.default_if_none(
            factory=default_system_title
        )  # type: ignore
    )

    global_encryption_key: Optional[bytes] = attr.ib(default=None)
    global_authentication_key: Optional[bytes] = attr.ib(default=None)

    # The security suite just indicates what cryptographical functions are available to
    # the meter
    security_suite: int = attr.ib(default=0)

    # Meter system title is usually unknown at the start of the connection. And it is
    # only needed if the connection is ciphered. In the AARE the meter will respond with
    # its system title
    meter_system_title: Optional[bytes] = attr.ib(default=None)

    # Meter authentication method.
    authentication_method: Optional[enums.AuthenticationMechanism] = attr.ib(
        default=None
    )
    # Low Level Security (LLS) password
    password: Optional[bytes] = attr.ib(default=None)

    # HLS challenge length.
    challenge_length: int = attr.ib(default=32)

    # client_to_meter_challenge is generated automatically with a random seed
    # depending on the HLS setup.
    client_to_meter_challenge: bytes = attr.ib(
        init=False,
        default=attr.Factory(
            lambda self: make_client_to_server_challenge(self.challenge_length),
            takes_self=True,
        ),
    )
    meter_to_client_challenge: Optional[bytes] = attr.ib(default=None, init=False)

    # To keep track of invocation counters used by the meter. If a request is received
    # with an invocation counter smaller than the current registered the message should
    # be rejected. If the meter receives a message with a with the same or lower
    # invocation counter as the last received message it will reject the message. So
    # after every use of the apdu protection the client_invocation_counter needs to be
    # increased
    client_invocation_counter: int = attr.ib(default=0)
    meter_invocation_counter: int = attr.ib(default=0)

    # Using dedicated ciphering increase security since we only use the global key for
    # association setup. This also saves a bit on the global invocation counter as once
    # it hits the maximum value the global key needs to be exchanged.
    use_dedicated_ciphering: bool = attr.ib(default=False)
    # Dedicated key will be generated on each association
    global_dedicated_key: Optional[bytes] = attr.ib(default=None)
    # the dedicated invocation_counter will be reset on each new dedicated_key.
    dedicated_invocation_counter: int = attr.ib(init=False, default=0)

    # not supported yet
    use_block_transfer: bool = attr.ib(default=False)

    # the max pdu size controls when we need to use block transfer. If the message is
    # larger than max_pdu_size we automatically use the general block service.
    # Unless it is not suppoeted in conformance. Then raise error.
    max_pdu_size: int = attr.ib(default=65535)

    # When a connection is preestablished we wont allow any ACSE adpus.
    is_pre_established: bool = attr.ib(default=False)

    buffer: bytearray = attr.ib(init=False, factory=bytearray)
    state: dlms_state.DlmsConnectionState = attr.ib(
        factory=dlms_state.DlmsConnectionState
    )

    conformance: Conformance = attr.ib(
        default=attr.Factory(
            lambda self: make_conformance(
                self.global_encryption_key, self.use_block_transfer
            ),
            takes_self=True,
        )
    )

    @classmethod
    def with_pre_established_association(
        cls,
        conformance: Conformance,
        max_pdu_size: int = 65535,
        global_encryption_key: Optional[bytes] = None,
        use_dedicated_ciphering: bool = False,
        client_system_title: Optional[bytes] = None,
    ):
        """
        A pre-established association does not need the ACSE APDUs. It is
        predetermined what access the client have.
        """
        return cls(
            client_system_title=client_system_title,
            global_encryption_key=global_encryption_key,
            # Moves the state into ready.
            state=dlms_state.DlmsConnectionState(current_state=dlms_state.READY),
            is_pre_established=True,
            conformance=conformance,
            max_pdu_size=max_pdu_size,
            use_dedicated_ciphering=use_dedicated_ciphering,
        )

    @property
    def security_control(self) -> security.SecurityControlField:
        """
        The security control field is used in encryption/decryption of data. It also
        follows the protected apdus to indicate what kind of protections they have.
        """
        _authenticated = bool(self.global_authentication_key)
        _encrypted = bool(self.global_encryption_key)
        return security.SecurityControlField(
            self.security_suite,
            encrypted=_encrypted,
            authenticated=_authenticated,
            broadcast_key=False,
        )

    @property
    def authentication_value(self) -> Optional[bytes]:
        """
        Depending on the authentication method for the connection the value in the
        authentication value of the AARQ is different.
        """
        if self.authentication_method is None:
            return None
        elif self.authentication_method == enums.AuthenticationMechanism.NONE:
            return None
        elif self.authentication_method == enums.AuthenticationMechanism.LLS:
            return self.password
        else:
            # HLS Mechanism
            return self.client_to_meter_challenge

    def send(self, event) -> bytes:
        """
        Returns the bytes to be sent over the connection and changes the state
         depending on event sent.
        :param event:
        :return: bytes
        """

        if self.is_pre_established:
            # When we are in a pre established association state starts as READY.
            # Only invalid state change is to send the ReleaseRequestApdu. But it is not
            # possible to close a pre-established association.
            if isinstance(event, acse.ReleaseRequestApdu):
                raise exceptions.PreEstablishedAssociationError(
                    f"You cannot send a {type(event)} when the association is"
                    f"pre-established "
                )

        self.state.process_event(event)
        LOG.debug(f"Client wants to send {event}")
        if self.use_protection:
            event = self.protect(event)

        # if self.use_blocks:
        #    blocks = self.make_blocks(event)
        #    # TODO: How to handle the subcase of sending blocks?
        LOG.info(f"Sending : {event}")
        return event.to_bytes()

    def receive_data(self, data: bytes):
        """
        Add data into the receive buffer.
        After this you could call next_event
        """
        if data:
            LOG.debug(f"Received DLMS data: {data!r}")
            self.buffer += data

    def next_event(self):
        """
        Will parse the buffer into an APDU. In lower levels we need the case to get more
        data. But this is not needed in the DLMS connections as it is the lower layers
        responisbility to make sure the data is complete before handing the control
        back to the dlms-layer. In the HDLC case the data is complete when we get the
        last nonsegmented informtion frame. And in the IP case we know the length of
        the data from the IP wrapper so we can keep on trying until we get all data.
        """
        apdu = XDlmsApduFactory.apdu_from_bytes(self.buffer)

        if self.use_protection:
            apdu = self.unprotect(apdu)

        self.update_negotiated_parameters(apdu)

        self.state.process_event(apdu)
        self.clear_buffer()

        if isinstance(apdu, acse.ApplicationAssociationResponseApdu):
            if apdu.result in [
                enums.AssociationResult.REJECTED_PERMANENT,
                enums.AssociationResult.REJECTED_TRANSIENT,
            ]:
                # reset the association on a reject
                self.state.process_event(dlms_state.RejectAssociation())

            # we need to start the HLS auth.
            elif apdu.authentication == enums.AuthenticationMechanism.HLS_GMAC:
                self.state.process_event(dlms_state.HlsStart())

        # Handle HLS verification
        if self.state.current_state == dlms_state.HLS_DONE:
            if isinstance(apdu, xdlms.ActionResponseNormalWithData):
                if apdu.status != enums.ActionResultStatus.SUCCESS:
                    self.state.process_event(dlms_state.HlsFailed())
                if self.hls_response_valid(utils.parse_as_dlms_data(apdu.data)):
                    self.state.process_event(dlms_state.HlsSuccess())
                else:
                    self.state.process_event(dlms_state.HlsFailed())
            elif isinstance(
                apdu, (xdlms.ActionResponseNormalWithError, xdlms.ActionResponseNormal)
            ):
                self.state.process_event(dlms_state.HlsFailed())

            else:
                raise exceptions.LocalDlmsProtocolError(
                    "Received a non Action response when in HLS DONE"
                )

        return apdu

    def clear_buffer(self):
        self.buffer = bytearray()

    @property
    def use_protection(self) -> bool:
        """
        If the Association is such that APDUs should be protected.
        :return:
        """
        if (
            self.global_encryption_key is not None
            or self.global_authentication_key is not None
        ):
            return True
        else:
            return False

    def protect(self, event) -> Any:
        """
        Will apply the correct protection to apdus depending on the security context
        """
        # ASCE have different rules about protection
        if isinstance(
            event, (acse.ApplicationAssociationRequestApdu, acse.ReleaseRequestApdu)
        ):
            # TODO: Not sure if it is needed to encrypt the IniateRequest when
            #   you are not sending a dedicated_key.
            if event.user_information:

                ciphered_initiate_text = self.encrypt(
                    event.user_information.content.to_bytes()
                )

                event.user_information = acse.UserInformation(
                    content=xdlms.GlobalCipherInitiateRequest(
                        security_control=self.security_control,
                        invocation_counter=self.client_invocation_counter,
                        ciphered_text=ciphered_initiate_text,
                    )
                )

        # XDLMS apdus should be protected with general-glo-cihpering
        elif isinstance(event, AbstractXDlmsApdu):
            ciphered_text = self.encrypt(event.to_bytes())
            LOG.info(f"Protecting a {type(event)} with GlobalCiphering")

            event = xdlms.GeneralGlobalCipherApdu(
                system_title=self.client_system_title,
                security_control=self.security_control,
                invocation_counter=self.client_invocation_counter,
                ciphered_text=ciphered_text,
            )
        else:
            raise RuntimeError(f"Unable to handle ecryption/protection of {event}")

        # updated the client_invocation_counter
        self.client_invocation_counter += 1
        return event

    def encrypt(self, plain_text: bytes):
        """
        Encrypts plain bytes according to the current association and connection.
        """
        if not self.global_encryption_key:
            raise ProtectionError(
                "Unable to encrypt plain text. Missing global_encryption_key"
            )
        if not self.global_authentication_key:
            raise ProtectionError(
                "Unable to encrypt plain text. Missing global_authentication_key"
            )

        return security.encrypt(
            self.security_control,
            system_title=self.client_system_title,
            invocation_counter=self.client_invocation_counter,
            key=self.global_encryption_key,
            auth_key=self.global_authentication_key,
            plain_text=plain_text,
        )

    def decrypt(self, ciphered_text: bytes):
        """
        Encrypts ciphered bytes according to the current association and connection.
        """

        if not self.global_encryption_key:
            raise ProtectionError(
                "Unable to decrypt ciphered text. Missing global_encryption_key"
            )
        if not self.global_authentication_key:
            raise ProtectionError(
                "Unable to decrypt ciphered text. Missing global_authentication_key"
            )
        if not self.meter_system_title:
            raise ProtectionError(
                "Unable to decrypt ciphered text. Have not received the meters system title."
            )

        return security.decrypt(
            self.security_control,
            system_title=self.meter_system_title,
            invocation_counter=self.meter_invocation_counter,
            key=self.global_encryption_key,
            auth_key=self.global_authentication_key,
            cipher_text=ciphered_text,
        )

    def unprotect(self, event):
        """
        Removes protection from APDUs and return a new the unprotected version
        """
        if isinstance(
            event, (acse.ApplicationAssociationResponseApdu, acse.ReleaseResponseApdu)
        ):
            if event.user_information:
                if isinstance(
                    event.user_information.content, xdlms.GlobalCipherInitiateResponse
                ):
                    received_invocation_counter = (
                        event.user_information.content.invocation_counter
                    )
                    self.validate_received_invocation_counter(
                        received_invocation_counter
                    )
                    self.meter_invocation_counter = received_invocation_counter
                    plain_text = security.decrypt(
                        security_control=event.user_information.content.security_control,
                        system_title=self.meter_system_title or event.system_title,
                        invocation_counter=event.user_information.content.invocation_counter,
                        key=self.global_encryption_key,
                        auth_key=self.global_authentication_key,
                        cipher_text=event.user_information.content.ciphered_text,
                    )
                    event.user_information.content = (
                        xdlms.InitiateResponseApdu.from_bytes(plain_text)
                    )

        elif isinstance(event, xdlms.GeneralGlobalCipherApdu):
            self.validate_received_invocation_counter(event.invocation_counter)
            self.meter_invocation_counter = event.invocation_counter
            plain_text = security.decrypt(
                security_control=event.security_control,
                system_title=event.system_title,
                invocation_counter=event.invocation_counter,
                key=self.global_encryption_key,
                auth_key=self.global_authentication_key,
                cipher_text=event.ciphered_text,
            )
            event = XDlmsApduFactory.apdu_from_bytes(plain_text)

        else:
            raise RuntimeError(f"Unable to handle decryption/unprotection of {event}")

        return event

    @property
    def use_blocks(self) -> bool:
        """
        If the event should be sent via GlobalBlockTransfer
        """
        return False

    def make_blocks(self, event) -> List[Any]:
        """
        Will split an APDU in blocks
        """
        return event

    def get_aarq(self) -> acse.ApplicationAssociationRequestApdu:
        """
        Returns an AARQ with the appropriate information for setting up the
        connection.
        """
        if self.global_encryption_key:
            ciphered_apdus = True
        else:
            ciphered_apdus = False

        initiate_request = xdlms.InitiateRequestApdu(
            proposed_conformance=self.conformance,
            client_max_receive_pdu_size=self.max_pdu_size,
        )

        return acse.ApplicationAssociationRequestApdu(
            ciphered=ciphered_apdus,
            system_title=self.client_system_title,
            authentication=self.authentication_method,
            authentication_value=self.authentication_value,
            user_information=acse.UserInformation(content=initiate_request),
        )

    def get_rlrq(self) -> acse.ReleaseRequestApdu:
        """
        Returns a ReleaseRequestApdu to release the current association.
        """
        initiate_request = xdlms.InitiateRequestApdu(
            proposed_conformance=self.conformance,
            client_max_receive_pdu_size=self.max_pdu_size,
        )

        return acse.ReleaseRequestApdu(
            reason=enums.ReleaseRequestReason.NORMAL,
            user_information=acse.UserInformation(content=initiate_request),
        )

    def update_negotiated_parameters(self, event: Any) -> None:
        """
        When an AARE is received we need to update the connection to the negotiated
        parameters from the server (meter)
        """
        if (
            self.state.current_state == dlms_state.AWAITING_ASSOCIATION_RESPONSE
            and isinstance(event, acse.ApplicationAssociationResponseApdu)
        ):
            if event.user_information:
                assert isinstance(
                    event.user_information.content, xdlms.InitiateResponseApdu
                )

                self.conformance = event.user_information.content.negotiated_conformance
                self.max_pdu_size = (
                    event.user_information.content.server_max_receive_pdu_size
                )
            self.meter_system_title = event.system_title
            self.authentication_method = event.authentication
            self.meter_to_client_challenge = event.authentication_value

    def get_hls_reply(self) -> bytes:
        """
        When the meter has enterted the HLS procedure the client firsts sends a reply
        to the server (meter) challenge. It is done with an ActionRequest to the
        current LN Association object in the meter. Method 2, Reply_to_HLS.

        Depending on the HLS type the data looks a bit different

        HLS_GMAC:
            SC + IC + GMAC(SC + AK + Challenge)
        """
        if not self.meter_to_client_challenge:
            raise exceptions.LocalDlmsProtocolError("Meter has not send challenge")
        if not self.global_encryption_key:
            raise ProtectionError(
                "Unable to create GMAC. Missing global_encryption_key"
            )
        if not self.global_authentication_key:
            raise ProtectionError(
                "Unable to create GMAC. Missing global_authentication_key"
            )
        if self.authentication_method == enums.AuthenticationMechanism.HLS_GMAC:
            only_auth_security_control = security.SecurityControlField(
                security_suite=self.security_suite, authenticated=True, encrypted=False
            )

            gmac_result = security.gmac(
                security_control=only_auth_security_control,
                system_title=self.client_system_title,
                invocation_counter=self.client_invocation_counter,
                key=self.global_encryption_key,
                auth_key=self.global_authentication_key,
                challenge=self.meter_to_client_challenge,
            )
            return (
                only_auth_security_control.to_bytes()
                + self.client_invocation_counter.to_bytes(4, "big")
                + gmac_result
            )
        else:
            raise NotImplementedError(
                f"No implementation for HSL: {self.authentication_method!r}"
            )

    def hls_response_valid(self, response_to_client_challenge: bytes) -> bool:
        """
        After sending the HLS reply to the meter the meter sends back the result of the
        client challenge in the ActionResponse. To make sure the meter has dont the HLS
        auth correctly we must validate the data.
        The data looks different depending on the HLS type

        HLS_GMAC:
            SC + IC + GMAC(SC + AK + Challenge)

        """

        security_control = security.SecurityControlField.from_bytes(
            response_to_client_challenge[0].to_bytes(1, "big")
        )
        invocation_counter = int.from_bytes(response_to_client_challenge[1:5], "big")
        gmac_result = response_to_client_challenge[-12:]

        if not self.global_encryption_key:
            raise ProtectionError(
                "Unable to verify GMAC. Missing global_encryption_key"
            )
        if not self.global_authentication_key:
            raise ProtectionError(
                "Unable to verify GMAC. Missing global_authentication_key"
            )
        if not self.meter_system_title:
            raise ProtectionError(
                "Unable to verify GMAC. Have not received the meters system title."
            )
        if not self.client_to_meter_challenge:
            raise ProtectionError(
                "Unable to verify GMAC. Have not received the meters system title."
            )

        correct_gmac = security.gmac(
            security_control=security_control,
            system_title=self.meter_system_title,
            invocation_counter=invocation_counter,
            key=self.global_encryption_key,
            auth_key=self.global_authentication_key,
            challenge=self.client_to_meter_challenge,
        )
        return gmac_result == correct_gmac

    def validate_received_invocation_counter(
        self, received_invocation_counter: int
    ) -> None:
        """
        The recevied invocation counter must be larger than the last one we registered.
        """
        if received_invocation_counter <= self.meter_invocation_counter:
            raise exceptions.LocalDlmsProtocolError(
                "Received invocation counter is not larger than the previous "
                "received one. "
            )



import logging
import socket
from typing import *

import attr

from dlms_cosem import exceptions
from dlms_cosem.protocol.wrappers import WrapperHeader, WrapperProtocolDataUnit

LOG = logging.getLogger(__name__)


@attr.s(auto_attribs=True)
class BlockingTcpTransport:
    """
    A TCP transport using Blocking I/O.
    """

    host: str
    port: int
    client_logical_address: int
    server_logical_address: int
    timeout: int = attr.ib(default=5)
    tcp_socket: Optional[socket.socket] = attr.ib(init=False, default=None)

    @property
    def address(self) -> Tuple[str, int]:
        return self.host, self.port

    def wrap(self, bytes_to_wrap: bytes) -> bytes:
        """
        When sending data over TCP or UDP it is necessary to wrap the data in the in
        the DLMS IP wrapper. This is so the server (meter) knows where the data is
        intended and how long the message is since there is no "final/poll" bit like
        in HDLC.
        """
        header = WrapperHeader(
            source_wport=self.client_logical_address,
            destination_wport=self.server_logical_address,
            length=len(bytes_to_wrap),
        )
        return WrapperProtocolDataUnit(bytes_to_wrap, header).to_bytes()

    def connect(self):
        """
        Create a new socket and set it on the transport
        """
        if self.tcp_socket:
            raise RuntimeError(f"There is already an active socket to {self.address}")

        try:
            self.tcp_socket = socket.create_connection(
                address=self.address, timeout=self.timeout
            )
        except (
            OSError,
            IOError,
            socket.timeout,
            socket.error,
            ConnectionRefusedError,
        ) as e:
            raise exceptions.CommunicationError("Unable to connect socket") from e
        LOG.info(f"Connected to {self.address}")

    def disconnect(self):
        """
        Close socket and remove it from the transport. No-op if the socket is already
        closed.
        """
        if self.tcp_socket:
            # only disconnect if there is a socket.
            try:
                self.tcp_socket.shutdown(socket.SHUT_RDWR)
                self.tcp_socket.close()
            except (OSError, IOError, socket.timeout, socket.error) as e:
                self.tcp_socket = None
                raise exceptions.CommunicationError from e
            self.tcp_socket = None
            LOG.info(f"Connection to {self.address} is closed")

    def send(self, bytes_to_send: bytes) -> bytes:
        """
        Sends a whole DLMS APDU wrapped in the DLMS IP Wrapper.
        """
        if not self.tcp_socket:
            raise RuntimeError("TCP transport not connected.")
        try:
            self.tcp_socket.sendall(self.wrap(bytes_to_send))
        except (OSError, IOError, socket.timeout, socket.error) as e:
            raise exceptions.CommunicationError("Could no send data") from e

        return self.recv()

    def recv(self) -> bytes:
        """
        Receives a whole DLMS APDU. Gets the total length from the DLMS IP Wrapper.
        """
        if not self.tcp_socket:
            raise RuntimeError("TCP transport not connected.")
        try:
            header = WrapperHeader.from_bytes(self.tcp_socket.recv(8))
            data = self.tcp_socket.recv(header.length)
        except (OSError, IOError, socket.timeout, socket.error) as e:
            raise exceptions.CommunicationError("Could not receive data") from e
        return data



from contextlib import contextmanager
from decimal import Decimal
from typing import *

import attr

from dlms_cosem import cosem, utils
from dlms_cosem.clients.dlms_client import DlmsClient
from dlms_cosem.cosem import Obis
from dlms_cosem.cosem.profile_generic import Data, ProfileGeneric
from dlms_cosem.protocol.xdlms.selective_access import RangeDescriptor


def force_obis(maybe_obis: Union[str, Obis]):
    if isinstance(maybe_obis, Obis):
        return maybe_obis
    else:
        return Obis.from_dotted(maybe_obis)


@attr.s(auto_attribs=True)
class NumericalMeterValue:
    """
    When parsing data from a DlmsMeter it can be a simple value that we know the meta
    data around before since it is predefined. For example static attribute on an
    interface class. For example capture_period on ProfileGeneric is an integer in
    minutes.
    Other attributes can be objects that is returned so that we may correctly parse
    data in other attributes: Ex. capture_objects in Profile Generic.
    Modeling values when they are not part of an attribute on an interface class is
    usually done via interface classes DATA and REGISTER.
    We also have numerial based values and string-based values.
    """

    # To avoid floating point rounding error we should always use Decimal to
    # represent a numerical value in a meter.
    value: Decimal

    # Scalar is used if the value provides one
    scalar: Optional[int]

    # see table 4 in Blue Book.
    # TODO: simple Enum or a factory that returns Unit objects?
    unit: Any


@attr.s(auto_attribs=True)
class StringlikeMeterValue:
    """
    For meter values that are string-like. Bytes should be represented as hexadecimal
    strings If we can't decode them to strings.
    """

    pass


@attr.s(auto_attribs=True)
class Meter:

    dlms_client: DlmsClient
    objects: Dict[str, Union[ProfileGeneric, Data]]

    def object_exists(self, object_obis: Obis) -> bool:
        return object_obis.dotted_repr() in self.objects.keys()

    @contextmanager
    def session(self):
        self.dlms_client.associate()
        yield self
        self.dlms_client.release_association()

    def get(
        self,
        logical_name: Union[str, Obis],
        attribute: int,
        selective_access: Optional[RangeDescriptor] = None,
    ):
        obis = force_obis(logical_name)
        instance = self.objects.get(obis.dotted_repr(), None)
        if instance is None:
            raise ValueError(
                f"Object with logical name {obis.dotted_repr()} does not exist on meter"
            )
        if instance.is_static_attribute(attribute):
            # check if the value is already present on the meter
            value = getattr(
                instance, instance.STATIC_ATTRIBUTES[attribute].attribute_name
            )
            if value:
                return value

        value = utils.parse_as_dlms_data(
            self.dlms_client.get(
                cosem_attribute=cosem.CosemAttribute(
                    interface=instance.INTERFACE_CLASS_ID,
                    instance=obis,
                    attribute=attribute,
                )
            )
        )

        if instance.is_static_attribute(attribute):
            setattr(
                instance, instance.STATIC_ATTRIBUTES[attribute].attribute_name, value
            )

        converter = instance.DYNAMIC_CONVERTERS.get(attribute, None)
        if converter:
            return converter(instance, value)
        return value



from dlms_cosem import a_xdr


def parse_as_dlms_data(data: bytes):
    data_decoder = a_xdr.AXdrDecoder(
        encoding_conf=a_xdr.EncodingConf(
            attributes=[a_xdr.Sequence(attribute_name="data")]
        )
    )
    return data_decoder.decode(data)["data"]



from typing import *

import attr

from dlms_cosem.a_xdr import (
    Attribute,
    AXdrDecoder,
    DlmsDataToPythonConverter,
    EncodingConf,
    Sequence,
)
from dlms_cosem.dlms_data import BaseDlmsData, DateTimeData
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu


@attr.s(auto_attribs=True)
class LongInvokeIdAndPriority:
    """
    Unsigned 32 bits

     - bit 0-23: Long Invoke ID
     - bit 25-27: Reserved
     - bit 28: Self descriptive -> 0=Not Self Descriptive, 1= Self-descriptive
     - bit 29: Processing options -> 0 = Continue on Error, 1=Break on Error
     - bit 30: Service class -> 0 = Unconfirmed, 1 = Confirmed
     - bit 31 Priority, -> 0 = normal, 1 = high.

    :param int long_invoke_id: Long Invoke ID
    :param bool self_descriptive: Indicates if self descriptive  `DEFAULT=False`
    :param bool confirmed: Indicates if confirmed. `DEFAULT=False`
    :param bool prioritized: Indicates if prioritized. `DEFAULT=False`
    :param bool break_on_error: Indicates id should break in error. `DEFAULT=True`

    """

    long_invoke_id: int
    prioritized: bool = attr.ib(default=False)
    confirmed: bool = attr.ib(default=False)
    self_descriptive: bool = attr.ib(default=False)
    break_on_error: bool = attr.ib(default=False)

    @classmethod
    def from_bytes(cls, bytes_data):
        if len(bytes_data) != 4:
            raise ValueError(
                f"LongInvokeIdAndPriority is 4 bytes long,"
                f" received: {len(bytes_data)}"
            )

        long_invoke_id = int.from_bytes(bytes_data[0:3], "big")
        status_byte = bytes_data[3]
        prioritized = bool(status_byte & 0b10000000)
        confirmed = bool(status_byte & 0b01000000)
        break_on_error = bool(status_byte & 0b00100000)
        self_descriptive = bool(status_byte & 0b00010000)

        return cls(
            long_invoke_id=long_invoke_id,
            prioritized=prioritized,
            confirmed=confirmed,
            break_on_error=break_on_error,
            self_descriptive=self_descriptive,
        )


@attr.s(auto_attribs=True)
class NotificationBody:
    """
    Sequence of DLMSData
    """

    ENCODING_CONF = EncodingConf(attributes=[Sequence(attribute_name="encoding_conf")])

    data: List[BaseDlmsData] = attr.ib(default=None)
    encoding_conf: EncodingConf = attr.ib(
        default=None
    )  # To store the data structure to be able to encode it again after initial decode.

    @classmethod
    def from_bytes(cls, bytes_data):
        decoder = AXdrDecoder(encoding_conf=cls.ENCODING_CONF)
        in_dict = decoder.decode(bytes_data)
        in_dict.update(
            {
                "data": DlmsDataToPythonConverter(
                    encoding_conf=in_dict["encoding_conf"]
                ).to_python()
            }
        )

        return cls(**in_dict)


@attr.s(auto_attribs=True)
class DataNotificationApdu(AbstractXDlmsApdu):
    """
    The DataNotification APDU is used by the DataNotification service.
    It is used to push data from a server (meter) to the client (amr-system).
    It is an unconfirmable service.

    A DataNotification APDU, if to large, can be sent using the general block
    transfer method.

    :param `LongInvokeAndPriority` long_invoke_id_and_priority: The long invoke
        id is a reference to the server invocation. self_descriptive,
        break_on_error and prioritized are not used for Datanotifications.
    :param datetime.datetime date_time: Indicates the time the DataNotification
        was sent. Is optional.
    :param `NotificationBody` notification_body: Push data.
    """

    TAG = 15
    NAME = "data-notification"

    ENCODING_CONF = EncodingConf(
        attributes=[
            Attribute(
                attribute_name="long_invoke_id_and_priority",
                create_instance=LongInvokeIdAndPriority.from_bytes,
                length=4,
            ),
            Attribute(
                attribute_name="date_time",
                create_instance=DateTimeData.from_bytes,
                optional=True,
                length=12,
            ),
            Sequence(
                attribute_name="notification_body",
                # create_instance=NotificationBody.from_bytes,
                # wrap_end=True,
            ),
        ]
    )

    # TODO: Verify if datetime has a length argument when sent. There is not
    #  set a specific length in the ASN.1 definition.
    #  so might be 0x01{length}{data}

    long_invoke_id_and_priority: LongInvokeIdAndPriority
    date_time: DateTimeData
    notification_body: NotificationBody

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        tag = bytes_data[0]
        if tag != cls.TAG:
            raise ValueError(f"Tag error. Expected tag {cls.TAG} but got {tag}")
        decoder = AXdrDecoder(encoding_conf=cls.ENCODING_CONF)
        in_dict = decoder.decode(bytes_data[1:])
        return cls(**in_dict)

    def to_bytes(self) -> bytes:
        raise NotImplementedError()



from typing import *

import attr


def allowed_range_for_obis_code(instance, attribute, value: int):

    if 0 > value > 255:
        raise ValueError("An obis can only be between 0 - 255")


@attr.s(auto_attribs=True)
class Obis:

    """
    As of now we ignore the special separators and focus on just dotted.
    """

    a: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code]
    )
    b: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code]
    )
    c: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code]
    )
    d: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code]
    )
    e: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code]
    )
    f: int = attr.ib(
        validator=[attr.validators.instance_of(int), allowed_range_for_obis_code],
        default=255,
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        if len(data) != 6:
            raise ValueError(
                f"Not enough data to parse OBIS. Need 6 bytes but got {len(data)}"
            )
        return cls(data[0], data[1], data[2], data[3], data[4], data[5])

    @classmethod
    def from_dotted(cls, dotted: str):
        numbers = dotted.split(".")
        if len(numbers) != 6:
            raise ValueError("dotted obis representation does not contain 6 elements")
        return cls(
            a=int(numbers[0]),
            b=int(numbers[1]),
            c=int(numbers[2]),
            d=int(numbers[3]),
            e=int(numbers[4]),
            f=int(numbers[5]),
        )

    def dotted_repr(self) -> str:
        return f"{self.a}.{self.b}.{self.c}.{self.d}.{self.e}.{self.f}"

    def verbose_repr(self) -> str:
        return f"{self.a}-{self.b}:{self.c}.{self.d}.{self.e}*{self.f}"

    def to_bytes(self) -> bytes:
        return bytes(bytearray([self.a, self.b, self.c, self.d, self.e, self.f]))



import logging
from typing import Optional

import attr
import serial

from dlms_cosem.hdlc import address, connection, frames, state

LOG = logging.getLogger(__name__)

LLC_COMMAND_HEADER = b"\xe6\xe6\x00"
LLC_RESPONSE_HEADER = b"\xe6\xe7\x00"


class ClientError(Exception):
    """General error in client"""


@attr.s(auto_attribs=True)
class SerialHdlcTransport:
    """
    HDLC transport to send data over serial.
    """

    client_logical_address: int
    server_logical_address: int
    serial_port: str
    serial_baud_rate: int = attr.ib(default=9600)
    server_physical_address: Optional[int] = attr.ib(default=None)
    client_physical_address: Optional[int] = attr.ib(default=None)
    hdlc_connection: connection.HdlcConnection = attr.ib(
        default=attr.Factory(
            lambda self: connection.HdlcConnection(
                self.server_hdlc_address, self.client_hdlc_address
            ),
            takes_self=True,
        )
    )
    _serial: serial.Serial = attr.ib(
        default=attr.Factory(
            lambda self: serial.Serial(
                port=self.serial_port, baudrate=self.serial_baud_rate, timeout=2
            ),
            takes_self=True,
        )
    )

    _send_buffer: list = attr.ib(factory=list)
    out_buffer: bytearray = attr.ib(init=False, factory=bytearray)
    in_buffer: bytearray = attr.ib(init=False, factory=bytearray)

    @property
    def server_hdlc_address(self):
        return address.HdlcAddress(
            logical_address=self.server_logical_address,
            physical_address=self.server_physical_address,
            address_type="server",
        )

    @property
    def client_hdlc_address(self):
        return address.HdlcAddress(
            logical_address=self.client_logical_address,
            physical_address=self.client_physical_address,
            address_type="client",
        )

    def connect(self):
        """
        Sets up the HDLC Connection by sending a SNRM request.

        """
        # TODO: Implement hdlc parameter negotiation in SNRM frame

        if self.hdlc_connection.state.current_state != state.NOT_CONNECTED:
            raise ClientError(
                f"Client tried to initiate a HDLC connection but connection state was "
                f"not in NOT_CONNECTED but in "
                f"state={self.hdlc_connection.state.current_state}"
            )
        snrm = frames.SetNormalResponseModeFrame(
            destination_address=self.server_hdlc_address,
            source_address=self.client_hdlc_address,
        )
        self.out_buffer += self.hdlc_connection.send(snrm)
        self.drain_out_buffer()
        ua_response = self.next_event()
        LOG.info(f"Received {ua_response!r}")
        return ua_response

    def disconnect(self):
        """
        Sends a DisconnectFrame
        :return:
        """
        disc = frames.DisconnectFrame(
            destination_address=self.server_hdlc_address,
            source_address=self.client_hdlc_address,
        )

        self.out_buffer += self.hdlc_connection.send(disc)
        self.drain_out_buffer()
        response = self.next_event()
        return response

    def next_event(self):
        """
        Will read the serial line until a proper response event is read.
        :return:
        """

        while True:
            # If we already have a complete event buffered internally, just
            # return that. Otherwise, read some data, add it to the internal
            # buffer, and then try again.
            event = self.hdlc_connection.next_event()
            if event is state.NEED_DATA:
                self.hdlc_connection.receive_data(self._read_frame())
                continue
            return event

    def send(self, telegram: bytes) -> bytes:
        """
        Send will make sure the data that needs to be sent i sent.
        The send is the only public function that will return the response data
        when received in full.
        Send will handle fragmentation of data if data is to large to be sent in a
        single HDLC frame.
        :param telegram:
        :return:
        """
        # prepend the LLC
        # The LLC should only be present in the first segmented information frame.
        # So instead we just prepend the data with it we know it will only be in the
        # intial information frame

        self.out_buffer += LLC_COMMAND_HEADER
        self.out_buffer += telegram
        self.drain_out_buffer()
        in_buffer = bytearray()
        while True:
            response = self.next_event()
            in_buffer += response.payload
            if response.segmented and response.final:
                # there is still data but server has send its max window size
                # tell the server to send more.
                rr = frames.ReceiveReadyFrame(
                    destination_address=self.server_hdlc_address,
                    source_address=self.client_hdlc_address,
                    receive_sequence_number=self.hdlc_connection.server_rsn,
                )
                self.out_buffer += self.hdlc_connection.send(rr)
                self.drain_out_buffer()

            if response.segmented and not response.final:
                # the server will send more frames.
                continue
            if not response.segmented and response.final:
                # this was the last frame
                break

        if not in_buffer.startswith(LLC_RESPONSE_HEADER):
            raise ValueError("The data is not prepended by the LLC response header")
        # don't return the LLC
        return in_buffer[3:]

    def drain_out_buffer(self):
        """
        If the data we need to send is longer than the allowed InformationFrame payload
        size we need to segment the data. It is done by splitting the payload into
        several frames and setting the segmented flag.
        To indicate that we are done with the sending a window we set final on the
        last I-frame. To indicated we are done sending all the data we set segmented to
        False
        :return:
        """
        data_size = self.hdlc_connection.max_data_size
        while len(self.out_buffer) > 0:
            data = self.out_buffer[:data_size]
            self.out_buffer = self.out_buffer[data_size:]
            segmented = bool(self.out_buffer)
            if self.hdlc_connection.state.current_state != state.IDLE:
                self._write_bytes(data)
                return
            # We dont handle window sizes so final is always true
            out_frame = self.generate_information_frame(
                data, segmented=segmented, final=True
            )
            self._write_frame(out_frame)
            # if it is the last frame we should not listen to possible RR frame
            if segmented:
                response = self.next_event()
                if isinstance(response, frames.ReceiveReadyFrame):
                    # send the next information frame.
                    continue
            break

    def generate_information_frame(
        self, payload: bytes, segmented: bool, final: bool
    ) -> frames.InformationFrame:
        return frames.InformationFrame(
            destination_address=self.server_hdlc_address,
            source_address=self.client_hdlc_address,
            payload=payload,
            send_sequence_number=self.hdlc_connection.server_ssn,
            receive_sequence_number=self.hdlc_connection.server_rsn,
            segmented=segmented,
            final=final,
        )

    def _write_frame(self, frame):
        frame_bytes = self.hdlc_connection.send(frame)
        LOG.info(f"Sending {frame!r}")
        self._write_bytes(frame_bytes)

    def _write_bytes(self, to_write: bytes):
        LOG.debug(f"Sending: {to_write!r}")
        self._serial.write(to_write)

    def _read_frame(self) -> bytes:
        in_bytes = self._serial.read_until(frames.HDLC_FLAG)

        if in_bytes == frames.HDLC_FLAG:
            # We found the first HDLC Frame Flag. We should read until the last one.
            in_bytes += self._serial.read_until(frames.HDLC_FLAG)

        return in_bytes

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()



# CRC CCITT - HDLC Style 16-bit
# In accordning with ANSI C12.18(2006)
# Using 0xFFFF as initial value
# Running over serial so all message bytes need to be reversed
# before calculation (because least significant bit is sent first)
# resulting crc bytes needs to be reversed to become in correct order
# The reversed crc is then XOR:ed with 0xFFFF
#
from ctypes import c_ushort
from typing import *


class CRCCCITT:
    crc_ccitt_table: List[int] = list()

    # The CRC's are computed using polynomials.

    crc_ccitt_constant = 0x1021

    def __init__(self):
        self.starting_value = 0xFFFF

        # initialize the pre-calculated tables
        if not len(self.crc_ccitt_table):
            self.init_crc_table()

    def calculate_for(self, input_data, lsb_first=False) -> bytes:
        """

        :param input_data:
        :param lsb_first: Indicate if the Least significant byte should be returned
            first (little endian)
        :return:
        """

        # need to revers bits in bytes
        reversed_data = reverse_byte_message(input_data)

        reversed_crc = self._calculate(reversed_data)
        lsb_rev = reversed_crc & 0x00FF
        lsb: int = ord(reverse_byte(lsb_rev))
        lsb ^= 0xFF
        lsb_byte = lsb.to_bytes(1, "big")
        msb_rev = (reversed_crc & 0xFF00) >> 8
        msb = ord(reverse_byte(msb_rev))
        msb ^= 0xFF
        msb_byte = msb.to_bytes(1, "big")

        if lsb_first:
            return b"".join([lsb_byte, msb_byte])
        else:
            return b"".join([msb_byte, lsb_byte])

    def _calculate(self, input_data: bytes):

        crc_value = self.starting_value

        for c in input_data:
            tmp = ((crc_value >> 8) & 0xFF) ^ c
            crc_shifted = (crc_value << 8) & 0xFF00
            crc_value = crc_shifted ^ self.crc_ccitt_table[tmp]

        return crc_value

    def init_crc_table(self):
        """The algorithm uses tables with pre-calculated values"""
        for i in range(0, 256):
            crc = 0
            c = i << 8

            for j in range(0, 8):
                if (crc ^ c) & 0x8000:
                    crc = c_ushort(crc << 1).value ^ self.crc_ccitt_constant
                else:
                    crc = c_ushort(crc << 1).value

                c = c_ushort(c << 1).value  # equivalent of c = c << 1

            self.crc_ccitt_table.append(crc)


def reverse_byte(byte_to_reverse):
    and_value = 1
    reversed_byte = 0
    for i in range(0, 8):
        reversed_byte += ((byte_to_reverse & and_value) >> i) * (2 ** (7 - i))
        and_value += and_value

    return (chr(reversed_byte)).encode("latin-1")


def reverse_byte_message(msg):
    reversed_mgs = b""
    for char in msg:
        reversed_mgs += reverse_byte(char)
    return reversed_mgs



import pytest

from dlms_cosem.hdlc import address, fields, frames


def test_hdlc_frame_format_field_from_bytes():
    in_bytes = bytes.fromhex("a01d")
    f = fields.DlmsHdlcFrameFormatField.from_bytes(in_bytes)
    assert f.length == 29
    assert not f.segmented


def test_hdlc_frame_format_filed_from_bytes_segmented():
    in_bytes = bytes.fromhex("a81d")
    f = fields.DlmsHdlcFrameFormatField.from_bytes(in_bytes)
    assert f.length == 29
    assert f.segmented


def test_hdlc_frame_format_raises_value_error_too_long():
    with pytest.raises(ValueError):
        fields.DlmsHdlcFrameFormatField(length=999999, segmented=False)


def test_hdlc_frame_format_to_bytes_not_segmented():
    f = fields.DlmsHdlcFrameFormatField(length=29, segmented=False)
    assert f.to_bytes().hex() == "a01d"


def test_hdlc_frame_format_to_bytes_segmented():
    f = fields.DlmsHdlcFrameFormatField(length=29, segmented=True)
    assert f.to_bytes().hex() == "a81d"


class TestHdlcAddress:
    def test_find_address(self):
        frame = "7ea87e210396a4090f01160002020f02160002020f03160002020f04160002020f05160002020f0616000204120017110009060000160000ff0202010902030f0116010002030f0216010002030f0316010002030f0416010002030f0516010002030f0616010002030f0716010002030f0816010002030f09160100016ff37e"
        frame_bytes = bytes.fromhex(frame)
        (
            destination_address,
            source_address,
        ) = address.HdlcAddress.find_address_in_frame_bytes(frame_bytes)
        assert destination_address == (16, None, 1)
        assert source_address == (1, None, 1)

    @pytest.mark.parametrize(
        "hdlc_address,resulting_bytes",
        [
            ((1, None, "client"), b"\x03"),
            ((16, None, "client"), b"\x21"),
            ((0b1001010, None, "server"), b"\x95"),
            ((0b0100101, None, "client"), b"\x4b"),
            ((1, 17, "server"), b"\x02\x23"),
        ],
        # TODO: need to find references of multiy byte addresses to test.
    )
    def test_client_address_just_logical(self, hdlc_address, resulting_bytes):
        add = address.HdlcAddress(
            logical_address=hdlc_address[0],
            physical_address=hdlc_address[1],
            address_type=hdlc_address[2],
        )
        assert add.to_bytes() == resulting_bytes


class TestCrc:
    def test_crc(self):

        data = "033f"
        correct_crc = "5bec"

        _crc = frames.HCS
        result = _crc.calculate_for(bytes.fromhex(data))
        assert result == bytes.fromhex(correct_crc)


class TestHdlcFrameValidation:
    def test_frame_is_enclosed_by_hdlc_flag(self):
        data = b"\x7effff\x7e"
        assert frames.frame_is_enclosed_by_hdlc_flags(data)

    @pytest.mark.parametrize("data", [b"\x7effff", b"ffff\x7e", b"ffff"])
    def test_frame_is_enclosed_by_hdlc_flag_fail(self, data):
        assert not frames.frame_is_enclosed_by_hdlc_flags(data)


class TestSnrmFrame:
    def test_parses_correct(self):
        out_data = bytes.fromhex("7ea00802232193bd647e")
        destination_address = address.HdlcAddress(
            logical_address=1, physical_address=17, address_type="server"
        )

        # Public client
        source_address = address.HdlcAddress(
            logical_address=16, physical_address=None, address_type="client"
        )
        snrm = frames.SetNormalResponseModeFrame(destination_address, source_address)

        assert snrm.to_bytes() == out_data


class TestUAFrame:
    def test_parser_correctly(self):
        out_data = bytes.fromhex(
            "7EA01F21022373E6C781801205019A06019A070400000001080400000001CCA27E"
        )
        # from device so public client is destination address.
        source_address = address.HdlcAddress(
            logical_address=1, physical_address=17, address_type="server"
        )

        # Public client
        destination_address = address.HdlcAddress(
            logical_address=16, physical_address=None, address_type="client"
        )

        information = out_data[9:-3]
        ua = frames.UnNumberedAcknowledgmentFrame(
            destination_address, source_address, information
        )
        print(ua.hcs.hex())
        print(ua.to_bytes().hex())
        print(out_data.hex())
        assert ua.to_bytes() == out_data

    def test_from_bytes(self):

        in_data = b"~\xa0\x1f!\x02#s\xe6\xc7\x81\x80\x12\x05\x01\x9a\x06\x01\x9a\x07\x04\x00\x00\x00\x01\x08\x04\x00\x00\x00\x01\xcc\xa2~"
        frame = frames.UnNumberedAcknowledgmentFrame.from_bytes(in_data)
        assert in_data == frame.to_bytes()


class TestInformationFrame:
    def test_construct(self):

        total = bytes.fromhex(
            "7EA02C02232110AF9FE6E600601DA109060760857405080101BE10040E01000000065F1F0400001E1DFFFFC5E47E"
        )
        information_part = bytes.fromhex(
            "E6E600601DA109060760857405080101BE10040E01000000065F1F0400001E1DFFFF"
        )
        server_address = address.HdlcAddress(
            logical_address=1, physical_address=17, address_type="server"
        )

        # Public client
        client_address = address.HdlcAddress(
            logical_address=16, physical_address=None, address_type="client"
        )
        frame = frames.InformationFrame(
            destination_address=server_address,
            source_address=client_address,
            payload=information_part,
            send_sequence_number=0,
            receive_sequence_number=0,
            segmented=False,
            final=True,
        )

        print(total)
        print(frame.to_bytes())
        assert frame.to_bytes() == total


class TestInformationResponseFrame:
    def test_contruct(self):
        total = bytes.fromhex(
            "7EA0382102233034E7E6E7006129A109060760857405080101A203020100A305A103020100BE10040E0800065F1F0400001E1D04C80007B86A7E"
        )
        information_part = bytes.fromhex(
            "E6E7006129A109060760857405080101A203020100A305A103020100BE10040E0800065F1F0400001E1D04C80007"
        )
        server_address = address.HdlcAddress(
            logical_address=1, physical_address=17, address_type="server"
        )

        # Public client
        client_address = address.HdlcAddress(
            logical_address=16, physical_address=None, address_type="client"
        )
        frame = frames.InformationFrame(
            destination_address=client_address,
            source_address=server_address,
            payload=information_part,
            send_sequence_number=0,
            receive_sequence_number=1,
            segmented=False,
            final=True,
        )

        print(f"length = {frame.frame_length}  :: {len(frame.to_bytes())}")

        print(total.hex())
        print(frame.to_bytes().hex())
        assert frame.to_bytes() == total

    def test_rebuild_frame(self):
        in_data = b"~\xa08!\x02#04\xe7\xe6\xe7\x00a)\xa1\t\x06\x07`\x85t\x05\x08\x01\x01\xa2\x03\x02\x01\x00\xa3\x05\xa1\x03\x02\x01\x00\xbe\x10\x04\x0e\x08\x00\x06_\x1f\x04\x00\x00\x1e\x1d\x04\xc8\x00\x07\xb8j~"
        info = frames.InformationFrame.from_bytes(in_data)
        print(info)
        assert info.to_bytes().hex() == in_data.hex()


class TestInformationControlField:
    def test_from_bytes(self):
        in_byte = bytes.fromhex("30")
        ctrl = fields.InformationControlField.from_bytes(in_byte)
        assert ctrl.receive_sequence_number == 1
        assert ctrl.send_sequence_number == 0
        assert ctrl.final



from dlms_cosem.protocol.acse.aare import ApplicationAssociationResponseApdu
from dlms_cosem.protocol.acse.aarq import ApplicationAssociationRequestApdu
from dlms_cosem.protocol.acse.base import *
from dlms_cosem.protocol.acse.rlre import ReleaseResponseApdu
from dlms_cosem.protocol.acse.rlrq import ReleaseRequestApdu
from dlms_cosem.protocol.acse.user_information import UserInformation

__all__ = [
    "ApplicationAssociationRequestApdu",
    "ApplicationAssociationResponseApdu",
    "ReleaseRequestApdu",
    "ReleaseResponseApdu",
    "AppContextName",
    "MechanismName",
    "UserInformation",
]
